/*
 * hst_opencl.c
 * This file is part of PyHST
 *
 * Copyright (C) 2011 - Matthias Vogelgesang
 * Copyright (C) 2002-2008 the European Synchrotron Radiation Facility (ESRF).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PyHST is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <CL/cl.h>

#include <clFFT.h>

extern "C" {
#include <glib.h>
}

#include "debug.h"
#include "hst_setup.h"
#include "hst_reconstructor.h"
#include "hst_opencl.h"

#define MAX_PROJECTIONS 4000

#define HST_KERNEL_FILTER       0
#define HST_KERNEL_BACKPROJECT  1
#define HST_KERNEL_SPREAD       2
#define HST_KERNEL_PACK         3
#define HST_KERNEL_NORMALIZE    4
/* NOTE: Update HST_KERNEL_SIZE when adding new kernels! */
#define HST_KERNEL_SIZE         5

/* NOTE: the following kernel names MUST match the preceeding IDs */
const char* HST_KERNEL_NAMES[HST_KERNEL_SIZE] = {
    "hst_filter",
    "hst_backproject",
    "hst_spread",
    "hst_pack",
    "hst_normalize_idft"
};

struct OpenCLContextT {
    /* This is one ugly hack... */
    HSTReconstructorContext hst_ctx;
#ifdef PYHST_MEASURE_TIMINGS
#define GPU_CONTEXT_MEMSET_OFFSET (3 * sizeof(GTimer*))
    GTimer *bp_timer;       //!< Counts time spent in backprojection code
    GTimer *filter_timer;    //!< Counts time spent in filtering
    GTimer *mem_timer;      //!< Counts time spent in memory operations
#else
#define GPU_CONTEXT_MEMSET_OFFSET 0
#endif

    int initialized;

    char                *device_name;
    cl_device_id        device_id;
    cl_context          clctx;
    cl_command_queue    queue;
    cl_program          program;
    cl_image_format     image_format;       /* TODO: We might drop this */
    cl_kernel           kernel[HST_KERNEL_SIZE];

    clFFT_Plan          fft_plan;

    cl_mem cosine_table;    /* pre-computed cos(alpha) values */
    cl_mem sine_table;      /* pre-computed sin(alpha) values */
    cl_mem axis_table;      /* axis correction values */
    cl_mem filter;          /* filter coefficients */
    cl_mem slice;           /* final result buffer */
    cl_mem sinogram;
    cl_mem buffer;          /* keeps FFT results */
    cl_mem frequencies;     /* keeps FFT results */
};

#ifdef PYHST_MEASURE_TIMINGS
#define hst_continue_timer(x) (g_timer_continue((x)))
#define hst_stop_timer(x) (g_timer_stop((x)))
#else
#define hst_continue_timer(x)
#define hst_stop_timer(x)
#endif


typedef struct OpenCLContextT OpenCLContext;

/*
 * Global variables for platform and device handling
 *
 * I have no idea if platform_ids and device_ids change after subsequent calls,
 * therefore we will store them during context creation in these global
 * structures and refer to them later on
 */
#define HST_MAX_OPENCL_DEVICES 16

struct {
    cl_platform_id platform_id;
    cl_uint num_devices;
    cl_device_id device_ids[HST_MAX_OPENCL_DEVICES]; 
} g_hst_platform;

static HSTConstString hst_opencl_timers[] = { "complete", "data transfer", "filtering", "backprojection", NULL }; 

/*
 * Utility functions
 */
const char* hst_opencl_error_msgs[] = {
    "CL_SUCCESS",
    "CL_DEVICE_NOT_FOUND",
    "CL_DEVICE_NOT_AVAILABLE",
    "CL_COMPILER_NOT_AVAILABLE",
    "CL_MEM_OBJECT_ALLOCATION_FAILURE",
    "CL_OUT_OF_RESOURCES",
    "CL_OUT_OF_HOST_MEMORY",
    "CL_PROFILING_INFO_NOT_AVAILABLE",
    "CL_MEM_COPY_OVERLAP",
    "CL_IMAGE_FORMAT_MISMATCH",
    "CL_IMAGE_FORMAT_NOT_SUPPORTED",
    "CL_BUILD_PROGRAM_FAILURE",
    "CL_MAP_FAILURE",
    "CL_MISALIGNED_SUB_BUFFER_OFFSET",
    "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST",

    /* next IDs start at 30! */
    "CL_INVALID_VALUE",
    "CL_INVALID_DEVICE_TYPE",
    "CL_INVALID_PLATFORM",
    "CL_INVALID_DEVICE",
    "CL_INVALID_CONTEXT",
    "CL_INVALID_QUEUE_PROPERTIES",
    "CL_INVALID_COMMAND_QUEUE",
    "CL_INVALID_HOST_PTR",
    "CL_INVALID_MEM_OBJECT",
    "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR",
    "CL_INVALID_IMAGE_SIZE",
    "CL_INVALID_SAMPLER",
    "CL_INVALID_BINARY",
    "CL_INVALID_BUILD_OPTIONS",
    "CL_INVALID_PROGRAM",
    "CL_INVALID_PROGRAM_EXECUTABLE",
    "CL_INVALID_KERNEL_NAME",
    "CL_INVALID_KERNEL_DEFINITION",
    "CL_INVALID_KERNEL",
    "CL_INVALID_ARG_INDEX",
    "CL_INVALID_ARG_VALUE",
    "CL_INVALID_ARG_SIZE",
    "CL_INVALID_KERNEL_ARGS",
    "CL_INVALID_WORK_DIMENSION",
    "CL_INVALID_WORK_GROUP_SIZE",
    "CL_INVALID_WORK_ITEM_SIZE",
    "CL_INVALID_GLOBAL_OFFSET",
    "CL_INVALID_EVENT_WAIT_LIST",
    "CL_INVALID_EVENT",
    "CL_INVALID_OPERATION",
    "CL_INVALID_GL_OBJECT",
    "CL_INVALID_BUFFER_SIZE",
    "CL_INVALID_MIP_LEVEL",
    "CL_INVALID_GLOBAL_WORK_SIZE"
};

static const char* hst_opencl_map_error(cl_int error)
{
    if (error >= -14)
        return hst_opencl_error_msgs[-error];
    if (error <= -30)
        return hst_opencl_error_msgs[-error-15];
}

#define CL_CHECK_ERROR(errorcode) if (errorcode != CL_SUCCESS) { \
    pyhst_error("OpenCL Error %i <%s:%i>", errorcode, __FILE__, __LINE__); }

#define CL_CHECK_ERROR_EX(errorcode) if (errorcode != CL_SUCCESS) { \
    pyhst_error("OpenCL Error: %s <%s:%i>", hst_opencl_map_error(errorcode), __FILE__, __LINE__); }


static int hst_cuda_configure_limits(OpenCLContext *ctx, HSTSetup *setup)
{
    float *param_s;

    int projection;
    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    float slope_zone = setup->pente_zone;

    float param;
    float axis_position_corr;
    float overlapping, flat_zone;

    param_s = (float*) malloc(2 * num_proj * sizeof(float));
    if (!param_s)
        return ENOMEM;

    for (projection = 0; projection < num_proj; projection++) {
        axis_position_corr = setup->axis_position_corr_s[projection];

        if (2 * axis_position_corr > num_bins) {
            overlapping = num_bins - axis_position_corr;
            param = 1; // pente_zone / prof_fact
        }
        else {
            overlapping = axis_position_corr;
            param = -1; // pente_zone / prof_fact
        }

        if (overlapping <= 0)
            return ERANGE;

        slope_zone = MIN(slope_zone, overlapping);
        flat_zone = overlapping - slope_zone;

        param_s[2*projection] = flat_zone;
        param_s[2*projection + 1] = param * slope_zone;
    }

    free(param_s);

    return 0;
}

/*
 * Implementation of HSTReconstructorContext interface
 */
static HSTReconstructorContext *hst_opencl_create_context(HSTReconstructor *prototype, HSTSetup *setup, int id)
{
    OpenCLContext *ctx;

    assert(prototype);
    assert(setup);
    assert(id < prototype->devices);

    ctx = (OpenCLContext*) malloc(sizeof(struct OpenCLContextT));
    if (ctx) {
        /* Obviously, this is poor man's polymorphism. */
        hst_reconstructor_init_context((HSTReconstructorContext*) ctx, prototype, setup);
        ctx->device_id = g_hst_platform.device_ids[id];

#ifdef PYHST_MEASURE_TIMINGS
        ctx->bp_timer = g_timer_new();
        if (ctx->bp_timer)
            g_timer_stop(ctx->bp_timer);
        ctx->filter_timer = g_timer_new();
        if (ctx->filter_timer)
            g_timer_stop(ctx->filter_timer);
        ctx->mem_timer = g_timer_new();
        if (ctx->mem_timer)
            g_timer_stop(ctx->mem_timer);

        if ((!ctx->bp_timer)||(!ctx->filter_timer)||(!ctx->mem_timer)) {
            if (ctx->bp_timer)
                g_timer_destroy(ctx->bp_timer);
            if (ctx->filter_timer)
                g_timer_destroy(ctx->filter_timer);
            if (ctx->mem_timer)
                g_timer_destroy(ctx->mem_timer);
            free(ctx);
            return NULL;
        }
#endif
    }
    ctx->initialized = 0;
    return (HSTReconstructorContext*) ctx;
}

static char* hst_opencl_load_program(const char* filename)
{
    FILE *fp = fopen(filename, "r");
    if (fp == NULL) {
        pyhst_error("Error: Couldn't open %s\n", filename);
        return NULL;
    }
    fseek (fp , 0 , SEEK_END);
    const size_t len = ftell(fp);
    rewind(fp);
    char *buffer;
    buffer = (char*) malloc (len * sizeof(char));
    if (buffer == NULL)
        return NULL;
    size_t buflen = fread(buffer, 1, len, fp);
    fclose(fp);
    if (buflen != len) {
        pyhst_error("Error: Couldn't read program %s from disk\n", filename);
        free(buffer);
        return NULL;
    }
    return buffer;
}

static int hst_opencl_init_context(HSTReconstructorContext *rctx, HWThread thrd)
{
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;
    assert(setup);

    assert(setup->num_projections < MAX_PROJECTIONS);
    int num_proj = setup->num_projections;
    int num_bins = setup->num_bins;
    int num_x = setup->num_x;
    int num_y = setup->num_y;

    ctx->device_name = (char *) malloc(256);
    snprintf(ctx->device_name, 256, "OpenCL %p", ctx->device_id);

    /* OpenCL context initialization */
    int err;
    ctx->clctx = clCreateContext(0, 1, &ctx->device_id, NULL, NULL, &err);
    CL_CHECK_ERROR(err);
    ctx->queue = clCreateCommandQueue(ctx->clctx, ctx->device_id, 0, &err);
    CL_CHECK_ERROR(err);

    char* source = hst_opencl_load_program("hst_opencl_kernels.cl");
    if (source == NULL)
        return 1; /* FIXME: find appropriate error number */

    ctx->initialized = 1;

    ctx->program = clCreateProgramWithSource(ctx->clctx, 1, (const char**) &source, NULL, &err);
    CL_CHECK_ERROR(err);
    free(source);

    err = clBuildProgram(ctx->program, 0, NULL, NULL, NULL, NULL);
    if (err != CL_SUCCESS) {
        const int LOG_SIZE = 4096;
        char* log = (char*) malloc(LOG_SIZE * sizeof(char));
        CL_CHECK_ERROR_EX(clGetProgramBuildInfo(ctx->program, ctx->device_id, CL_PROGRAM_BUILD_LOG, LOG_SIZE, (void*) log, NULL));
        pyhst_error("OpenCL Build Log\n---------------\n%s", log);
        free(log);
        return 255;
    }

    int kernel_id = 0;
    for (; kernel_id < HST_KERNEL_SIZE; kernel_id++) {
        ctx->kernel[kernel_id] = clCreateKernel(ctx->program, HST_KERNEL_NAMES[kernel_id], &err);
        CL_CHECK_ERROR_EX(err);
        if (err != CL_SUCCESS)
            printf("Undefined kernel name %s\n", HST_KERNEL_NAMES[kernel_id]);
    }

    /* Allocate and copy buffers, we don't use immediate copying during buffer
     * creation, since we might copy the data to all devices. */
    hst_continue_timer(ctx->mem_timer);

    ctx->cosine_table = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY, num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->cosine_table, CL_TRUE, 0, num_proj*sizeof(float), setup->cos_s, 0, NULL, NULL));

    ctx->sine_table = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY, num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->sine_table, CL_TRUE, 0, num_proj*sizeof(float), setup->sin_s, 0, NULL, NULL));

    ctx->axis_table = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY, num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->axis_table, CL_TRUE, 0, num_proj*sizeof(float), setup->axis_position_corr_s, 0, NULL, NULL));

    ctx->filter = clCreateBuffer(ctx->clctx, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 2*setup->dim_fft*sizeof(float), setup->filter, &err);
    CL_CHECK_ERROR(err);

    ctx->buffer = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, 2*setup->dim_fft*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->slice = clCreateBuffer(ctx->clctx, CL_MEM_WRITE_ONLY, num_x*num_y*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    ctx->sinogram = clCreateBuffer(ctx->clctx, CL_MEM_READ_WRITE, num_bins*num_proj*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    hst_stop_timer(ctx->mem_timer);

    /* Create FFT plan */
    clFFT_Dim3 dim;
    dim.x = setup->dim_fft; dim.y = 1; dim.z = 1;
    ctx->fft_plan = clFFT_CreatePlan(ctx->clctx, dim, clFFT_1D, clFFT_InterleavedComplexFormat, &err);
    CL_CHECK_ERROR(err);

    /* Set the arguments here, to avoid further calls */
    cl_kernel kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 3, sizeof(cl_mem), (void *) &ctx->cosine_table));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 4, sizeof(cl_mem), (void *) &ctx->sine_table));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 5, sizeof(cl_mem), (void *) &ctx->axis_table));

    return 0;
}

static int hst_opencl_configure(HSTReconstructorContext *rctx, HSTChanged changes)
{
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;

    if (changes & HST_FILTER_CHANGED) {
        hst_continue_timer(ctx->mem_timer);
        CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->filter, CL_TRUE, 0, 2*setup->dim_fft*sizeof(float), setup->filter, 0, NULL, NULL));
        hst_stop_timer(ctx->mem_timer);
    }

    if (changes & HST_LIMITS_CHANGED) {
        if (setup->fai360) {
            /* FIXME: check some obscure limits */
        }
    }
    return 0;
}

static void hst_opencl_free_context(HSTReconstructorContext *rctx)
{
    OpenCLContext *ctx;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;

    if (ctx->initialized) {
        free(ctx->device_name);
        clFFT_DestroyPlan(ctx->fft_plan);
        CL_CHECK_ERROR(clReleaseMemObject(ctx->cosine_table));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->sine_table));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->axis_table));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->filter));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->slice));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->sinogram));
        CL_CHECK_ERROR(clReleaseMemObject(ctx->buffer));
        CL_CHECK_ERROR(clReleaseCommandQueue(ctx->queue));
        int kernel_id = 0;
        for (; kernel_id < HST_KERNEL_SIZE; kernel_id++) {
            CL_CHECK_ERROR(clReleaseKernel(ctx->kernel[kernel_id]));
        }
        CL_CHECK_ERROR(clReleaseContext(ctx->clctx));
    }

    memset(((char*) ctx) + sizeof(HSTReconstructorContext), 0, sizeof(OpenCLContext) - sizeof(HSTReconstructorContext));
}

static void hst_opencl_destroy_context(HSTReconstructorContext *rctx)
{
#ifdef PYHST_MEASURE_TIMINGS
    OpenCLContext *ctx = (OpenCLContext *) rctx;
    if (ctx->filter_timer)
        g_timer_destroy(ctx->filter_timer);
    if (ctx->bp_timer)
        g_timer_destroy(ctx->bp_timer);
    if (ctx->mem_timer)
        g_timer_destroy(ctx->mem_timer);
#endif

    hst_opencl_free_context(rctx);
    hst_reconstructor_free_context(rctx);
    free(rctx);
}

static HSTConstString hst_opencl_get_title(const HSTReconstructorContext *rctx)
{
    assert(rctx);
    return ((OpenCLContext *) rctx)->device_name;
}

static HSTConstString *hst_opencl_get_timers(HSTReconstructorConstContextPtr rctx, double *timers) {
#ifdef PYHST_MEASURE_TIMINGS
    assert(rctx);
    
    if (timers) {
        OpenCLContext *ctx = (OpenCLContext *) rctx;
        timers[1] = g_timer_elapsed(ctx->mem_timer, NULL);
        timers[2] = g_timer_elapsed(ctx->filter_timer, NULL);
        timers[3] = g_timer_elapsed(ctx->bp_timer, NULL);
        timers[0] = timers[1] + timers[2] + timers[3];
    }
#endif

    return hst_opencl_timers;
}

static int hst_opencl_reconstruct(HSTReconstructorContext *rctx, float* slice, const float *sinogram)
{
    OpenCLContext *ctx;
    HSTSetup *setup;

    assert(rctx);
    ctx = (OpenCLContext*) rctx;
    setup = rctx->setup;
    assert(setup);
    assert(slice);

    int dim_fft = setup->dim_fft;
    int num_bins = setup->num_bins;
    int num_proj = setup->num_projections;
    int num_x = setup->num_x;
    int num_y = setup->num_y;
    int mid = (num_bins + dim_fft) / 2;
    float axis_position = setup->axis_position;
    cl_kernel kernel;
    size_t global_work_size[2];
    size_t local_work_size[2];

    cl_int err = CL_SUCCESS;

    /* 1. Copy sinogram data onto device */
    hst_continue_timer(ctx->mem_timer);
    CL_CHECK_ERROR(clEnqueueWriteBuffer(ctx->queue, ctx->sinogram, CL_TRUE, 0, num_bins*num_proj*sizeof(float), sinogram, 0, NULL, NULL));
    hst_stop_timer(ctx->mem_timer);

    /* 2. Do FFT */
    hst_continue_timer(ctx->filter_timer);
    if (dim_fft > num_bins) {
        if (setup->zero_padding) {
        }
        else {
            kernel = ctx->kernel[HST_KERNEL_SPREAD];
            global_work_size[0] = dim_fft;
            global_work_size[1] = num_proj;
            CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
            CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->sinogram));
            CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(int), &num_bins));
            CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 2, NULL, global_work_size, NULL, 0, NULL, NULL));
        }
    }
    else {
    }

    CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queue, ctx->fft_plan, num_proj, clFFT_Forward, ctx->buffer, ctx->buffer, 0, NULL, NULL));

    /* 3. Filter */
    kernel = ctx->kernel[HST_KERNEL_FILTER];
    global_work_size[0] = 2*dim_fft;
    global_work_size[1] = num_proj;
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->filter));
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 2, NULL, global_work_size, NULL, 0, NULL, NULL));

    /* 4. Do inverse FFT */
    CL_CHECK_ERROR(clFFT_ExecuteInterleaved(ctx->queue, ctx->fft_plan, num_proj, clFFT_Inverse, ctx->buffer, ctx->buffer, 0, NULL, NULL));

    /*
    kernel = ctx->kernel[HST_KERNEL_NORMALIZE];
    global_work_size[0] = dim_fft*num_proj;
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 1, NULL, global_work_size, NULL, 0, NULL, NULL));
    */

    kernel = ctx->kernel[HST_KERNEL_PACK];
    global_work_size[0] = dim_fft;
    global_work_size[1] = num_proj;
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *) &ctx->buffer));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(int), &num_bins));
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 2, NULL, global_work_size, NULL, 0, NULL, NULL));

    hst_stop_timer(ctx->filter_timer);

    /* 5. Back-project */
    hst_continue_timer(ctx->bp_timer);

    float off_x = setup->offset_x - setup->axis_position - 0.5f;
    float off_y = setup->offset_y - setup->axis_position - 0.5f;
    kernel = ctx->kernel[HST_KERNEL_BACKPROJECT];
    CL_CHECK_ERROR(clSetKernelArg(kernel, 0, sizeof(int), &num_proj));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 1, sizeof(int), &off_x));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 2, sizeof(int), &off_y));
    /* Parameters 3-5 are already set */
    CL_CHECK_ERROR(clSetKernelArg(kernel, 6, sizeof(cl_mem), (void *) &ctx->sinogram));
    CL_CHECK_ERROR(clSetKernelArg(kernel, 7, sizeof(cl_mem), (void *) &ctx->slice));

    /* TODO: Currently, we let OpenCL decide on how to decompose the work. This
     * could lead to performance degradations. */
    global_work_size[0] = num_x;
    global_work_size[1] = num_y;
    local_work_size[0] = 16;
    local_work_size[1] = 16;

    /* FIXME: In this case num_x = num_bins, but this is not necessarily always
     * the case! */
    CL_CHECK_ERROR(clEnqueueNDRangeKernel(ctx->queue, kernel, 2, NULL, global_work_size, local_work_size, 0, NULL, NULL));

    hst_stop_timer(ctx->bp_timer);

    /* 6. Put result back into slice buffer */
    hst_continue_timer(ctx->mem_timer);
    CL_CHECK_ERROR(clEnqueueReadBuffer(ctx->queue, ctx->slice, CL_TRUE, 0, num_x*num_y*sizeof(float), slice, 0, NULL, NULL));
    hst_stop_timer(ctx->mem_timer);

    return 0;   /* XXX: What does the return value do? */
}

static HSTReconstructor hst_opencl_info = {
    0,
    hst_opencl_get_title,
    hst_opencl_create_context,
    hst_opencl_init_context,
    hst_opencl_free_context,
    hst_opencl_destroy_context,
    hst_opencl_configure,
    NULL,   /* no pre-processing */
    hst_opencl_reconstruct,
    hst_reconstructor_postprocess_slice,
    hst_opencl_get_timers
};

/*
 * Implementation of public interface
 */
HSTReconstructor *hst_opencl_init()
{
    CL_CHECK_ERROR(clGetPlatformIDs(1, &(g_hst_platform.platform_id), NULL));

    clGetDeviceIDs(g_hst_platform.platform_id,
                CL_DEVICE_TYPE_GPU,
                HST_MAX_OPENCL_DEVICES,
                g_hst_platform.device_ids,
                &(g_hst_platform.num_devices));

    hst_opencl_info.devices = g_hst_platform.num_devices;
    return &hst_opencl_info;
}

void hst_opencl_free()
{
}


